/* 
scan.c -- egyszeru parser

Szimp hazi
Zsban Ambrus


BEVEZETO:

Ez egy egy egyszeru parser es interpreter.  Meg sajnos befejezetlen,
sok dolog hianyzik belole.  Flexben es c++-ban irodott.
Igyekeztem minnel tobb c++-os featuret hasznalni benne.

Elonyei:

 Letre lehet hozni es meghivni felhasznaloi fuggvenyeket.  
 
 Akar rekurzivan is.
 
 Majdnem valodi lexikai lokalis valtozokat kezel.
 Closure-okatt nem lehet letrehozni a szokasos modon, de tobb fuggvenyhez
 tartozo lokalis valtozot igen.
 
Hatranyai:

 Nincs benne hibakezeles, ehhez nagyon sok pointert auto_ptr-re kene
 atirni, es meg egyeb atalakitasokat is kene csinalni.  Minden hibara egy --
 altalaban primitiv -- hibauzenetet ir ki, es kilep.

 Valoszinuleg vannak meg benne memory leakek.  Meg kene keresni oket.
 Lehet, hogy komolyabb bajok is vannak benne.

 Referenciaszamlalos gc van benne, ezert a ciklikus szerkezeteket mint
 (var f={f});
 nem szabaditja fel.

 Nincs jo dokuja

 A fuggvenyeknek meg nem lehetnek parameterei.

 Nincs ++increment--, bitwise es shift operatorok

 Meg nem letezik a tomb tipus.

 Nagyon keves beepitett fuggveny van.
 
 Meg nem lehet stringeket kiirni.
 
 Nem lehet closure-okat csinalni.
 
 A balertek csak valtozo lehet, felteteles kif nem

 Nincsenek source level debugging funkciok.

 Valtozok elmentesevel kapcsolatos koncepcios hiba, lasd a megjegyzest
 az expr_newvar-nal.

Mivel sok hianyossag van, ezek kozul egyik-masikbol a beszamolon feladatot
csinalhatsz.  Persze nem ezert van sok hianyossag.


FORDITANI:

cp scan.hxx scan.hh
g++ -O2 -g -Wall -c -o scan.o scan.cxx
flex  -olex.yy.cc scan.l
g++ -O2 -g -c -o lex.yy.o lex.yy.cx
g++ -g -o scan scan.o lex.yy.o

Ahol -O2 helyettesitheto erosebb optimizaciokkal.
Ha igy nem megy, ertelemszeruen korrigald.

NYELV:

Harom tipus van (jelenleg, kesobb lehet tomb is):

 A void tipus, amit a void fuggvenyek adnak vissza, a valtozok
 alapertelmezett kezdoerteke.  A python None-jahoz, vagy a perl 
 undef-jehez hasonlo.

 Az egesz, ami egy C longot tarol.

 A lambda, ami egy fuggvenyt tarol, amit kesobb meg lehet hivni.
 
Interaktiv modban promptot adunk, es pontosvesszovel elvalasztott
kifejezeseket olvasunk be es hajtunk vegre.  A pontosvesszo utani ujsor
inditja a kiertekelest.  Ezt miattad csinaltam igy.  Ilyenkor az 'a'..'z'
valtozok elore deklaralva vannak.  Az egesz visszateresi erteku utasitasok
eredmenyet automatikusan kiirja.

Nem-interaktiv modban egyben beolvassuk az input utasitast, es csak a vegen
hajtjuk vegre.

Megjegyzest '#'-sel szurhatunk be, a sor vegeig tart.
A program neha debug informaciokat ir ki, ezek '#'-kal kezdodnek, es
altalaban nem kell foglalkozni veluk.

A szintaxis es a precedencia meglehetosen szokatlan.

Egy kifejezes a kovetkezo lehet (feljebb a magasabb precedenciajuak)
(remelem, nem hagyok ki semmit):

 Decimalis egyesz literal
 
 Elozoleg deklaralt valtozonev
 
 Valtozo-deklaracio, ha mar van ilyen nevu valtozo, akkor azt atfedi;
 ha ugyanebben a scope-ban van, akkor orokre: 'var' nev
 
 Beepitett fuggvey hivasa, ha nincs parametere a fgv-nek, akkor a zarojelek
 nem kotelezoek: beepitett-fuggveny-neve [ '(' [ kif [ ',' kif ] ] ')' ]
  
 Zarojeles kif: '(' esetleg ures utasitas ')'

 Ciklus kif, vegtelen ciklusban ismetli a tartalmat: 
 '(' esetleg ures utasitas ';' '*' ')'

 Lambda literal, lambdat ad vissza: '{' esetleg ures utasitas '}'

 Az elozo harmat blokknak hivom. Mindharom (es csak ezek) lexikai scopeot
 alkotnak, tehat az ilyenekben deklaralt valtozok neve kivulrol nem
 elerheto, de maga a valtozo esetleg meg letezhet.
 
 Lambda hivasa: kif '(' ')'

 A lambda futasideju scope-ot is letrehoz, vagyis elmentjuk az 
 osszes lokalis valtozot, hogy a reqrzio lehetseges legyen.

 Egesz aritmetika: kif ('*'|'/'|'%') kif

 Egesz aritmetika, vegyuk eszre, hogy az egyoperandusu elojeleknek is
 alacsonyabb a precedenciaja, mint a szorzasnak.  Ez altalaban mindegy, mert
 -(a*b)==(-a)*b, de -(f(x))<>(-f)(x): (kif| ) ('+'|'-') kif  
 
 Osszehassonlitas, 0=igaz, 1=hamis: kif ('=='|'<>'|'<'|'>'|'<='|'>=')
 
 Logikai nem, 0->1, nem0->0: '^' kif 
 
 Logikai es, ha az elso arg nem0, a masodikat adja, kulonben az elsot:
 kif '&&' kif
 
 Logikai vagy, ha az elso arg 0, a masodikat adja, kulonben az elsot:
 kif '||' kif
 
 Ertekadas, esetleg aritmetikai muvelettel egyutt:
 valtozo ('='|'+='|'-='|'*='|'/='|'%=') kif

 Progn, mint a C vesszzo operatora: kif '::' kif
 
 Logikai es, ugy mint &&, if utasitas helyett hasznalhato: kif '??' kif
 Logikai vagy, ugy mint ||, if(!) utasitas helyett, '??'-vel azonos 
 precedenciaju: kif '!!' kif
  
Utasitasok:

 Kifejezes
 
 Ures utasitas: 
 
 Egymasutan, foleg '?' vagy '!' belsejeben hasznos: utasitas ':' utasitas
 
 Egymasutan, a kozonseges parancselvalaszto: utasitas ';' utasitas
 
 Felteteles kilepes, ha a feltetel igaz, kilep a legbelso blokkbol, 
 es a blokk az utasitast adja vissza.  Ciklusbol valo kilepesre, vagy a
 nagyon gyakori "if (...) return (...)" szerkezetre, vagy "a?b:c" felteteles
 kifejezesnek jo. Az elso pontosvesszo hatarolja az utasitast:
 feltetel '?' utasitas_ha_igaz ';' esetleges tovabbi_utasitasok
 
 Forditott felteteles kifejezes, ugyanaz, mint az elozo, de a feltetel
 forditva ertendo. Tehat C "while(a){b}" => "(a!;b;);".
 A precedenciaja a '?'-lel megegyezo:
 feltetel '!' utasitas_ha_hamis ';' esetleges tovabbi_utasitasok
 
Beepitett fuggvenyek:

 Kilepes: exit
 
 Kiiratas, ha nem egeszre hivod meg, nem csinal semmit; egeszt kiirja, es
 newlinet rak utana: print (x)
 
 abs (x)
 
 Rand egesz [0..n)-bol; a random seed automatikusan be van allitva a
 progi inditasakor: rand (n)
 
 Minimum/maximum: min (a, b); max (a, b)
 
 Tipus ellenorzese: voidp (x); intp (x); lambdap (x)
 
PELDAK:

# egyszeru primkereso (nem igazan hatekony)
time echo 'var a=1;var x;(a+=1;x=1;(x+=1;x*x<=a!1;a%x!0;*)??
	print(a);a<110!;*)' | ./scan

# lexical scoping: a ket 'x' teljesen fuggetlen
echo 'var x=123456;var k={x/=2};var a=(var x=1;{print(x+=1);print(k());});
	var i=0; (i+=1;i>6?;print(a());*);print(x);'| ./scan

# a rekurziv fuggvenyek uj peldanyt kapnak lokalis valtozoikbol, de ebben
# az eggyel kiljebbi ertek van:
echo 'var f={var x;x=(voidp(x)?0;x)+1;print(x);rand(6)??f();print(x);};
	f(); f();' | ./scan 

s> # lexical scoping (namespace-ek)
s> (var x=-1; a= {x*=2}; b= {x/=2};);
s> (var x=1; c= {x+=1};);
s> x= 19; c(); a(); c(); a(); c(); b(); x;
19
2
-2
3
-4
4
-2
19
s> exit



VEGE a dokumentacionak, kezdodik a progi

*/


#include "scan.hh"

#include <algorithm>	// min
#include <cerrno>	// errno
#include <climits>	// LONG_MAX
#include <cstdlib>	// exit
#include <iostream>	// cout
#include <map>		// map
#include <stack>	// stack
#include <string>	// string
#include <ctime>	// time
#include <unistd.h>	// isatty

using std::cerr;
using std::cout;
using std::stack;
using std::min;
using std::max;
using std::map;



// elore deklaraciok -------------------------------------------------
// a korok elkerulesere

class Expr;


// Xref ------------------------------------------------------------
// hivatkozasszamlalos mutato

template <class C> class Xref {	
	public:
	Xref<C> *newref (void);
	void delref (void);
	Xref (C ***);
	Xref (C *);
	C *get (void);
	C **adr (void);
	private:
	C *a;
	int rc;
	};
template <class C> Xref<C>::Xref (C ***dp):
	a (0), rc (1) { *dp= &a; }
template <class C> Xref<C>::Xref (C *u):
	a (u), rc (1) { }
template <class C> Xref<C> *Xref<C>::newref (void)
	{ rc++; return this; }
template <class C> void Xref<C>::delref (void) {
	rc--;
	if (!rc) {
		cerr <<"#[DXref]\n"; // <---- DEBUG
		delete a;
		delete this; // jaj
		}
	}
template <class C> C *Xref<C>::get (void)
	{ return a; }
template <class C> C **Xref<C>::adr (void)
	{ return &a; }


// hibakezeles -------------------------------------------------------

void fatal (const char *s) {
	cerr <<s <<'\n';
	exit (1);
	};



// Val ------------------------------------------------------
// Itt igazabol dynamic_cast-et kene hasznalni isa fgv helyett, de
// a gcc 2.95 meg nem tamogatja.  Majd ha lesz ujabb gcc-m, atirom.


class Val {
	public:
	virtual ~Val (void);
	virtual int isa (int) const= 0;
	virtual Val *dup (void) const= 0;
	protected:
	Val (void);
	};
Val::Val (void) { };
Val::~Val (void) { };
	
class Val_int: public Val {
	public:
	Val_int (long);
	long a;
	int isa (int n) const { return n&type; }
	static const int type= 1; 
	Val *dup (void) const;
	};
Val_int::Val_int (long t):
	a (t) { };
Val *Val_int::dup (void) const
	{ return new Val_int (*this); }
	
class Val_void: public Val {
	public:
	Val_void (void);
	int isa (int n) const { return n&type; }
	static const int type= 2;
	Val *dup (void) const;
	};
Val_void::Val_void (void)
	{ }
Val *Val_void::dup (void) const
	{ return new Val_void; }

class Val_loop: public Val {
	public:
	Val_loop (void);
	int isa (int n) const { return n&type; }
	static const int type= 4;
	Val *dup (void) const;
	};
Val_loop::Val_loop (void)
	{ }
Val *Val_loop::dup (void) const
	{ fatal ("internal error: Val_loop dup'ed"); return 0; }

class Val_lambda: public Val {
	public:
	Val_lambda (Xref<Expr> *);
	Xref<Expr> *get (void);
	~Val_lambda (void);
	int isa (int n) const { return n&type; }
	static const int type= 8; 
	Val *dup (void) const;
	private:
	Xref<Expr> *a;
	};
Val_lambda::Val_lambda (Xref<Expr> *t):
	a (t) {
	a->newref ();
	};
Val *Val_lambda::dup (void) const
	{ return new Val_lambda (a); }
Xref<Expr> *Val_lambda::get (void)
	{ return a; }



// Expr --------------------------------------------------


class Expr {
	public:
	virtual void print (void) const= 0;
	virtual Val *eval (void) const;
	virtual Val **leval (void) const;
	virtual int islvalue (void) const;
	virtual ~Expr ();
	protected:
	Expr (void);
	};


void Expr::print (void) const 
	{ fatal ("internal error: abstract Expr"); }
Val *Expr::eval (void) const 
	{ fatal ("internal error: abstract Expr"); return 0; }
Val **Expr::leval (void) const 
	{ fatal ("internal error: assignment to non-lvalue"); return 0; }
int Expr::islvalue (void) const 
	{ return 0; }
Expr::~Expr (void) { }	
Expr::Expr (void) { } 

class Expr_void: public Expr {
	public:
	void print (void) const;
	Val *eval (void) const;
	Expr_void (void);
	};
void Expr_void::print (void) const
	{ cout <<"void"; }
Val *Expr_void::eval (void) const
	{ return new Val_void; }
Expr_void::Expr_void (void)
	{ }

class Expr_loop: public Expr {
	public:
	void print (void) const;
	Val *eval (void) const;
	Expr_loop (void);
	};
void Expr_loop::print (void) const
	{ cout <<"loop"; }
Val *Expr_loop::eval (void) const
	{ return new Val_loop; }
Expr_loop::Expr_loop (void)
	{ }

class Expr_num: public Expr {
	public:
	void print (void) const;
	Val *eval (void) const;
	Expr_num (long l);
	private:
	const long a;
	};
Expr_num::Expr_num (long l):
	a (l) { }
void Expr_num::print (void) const
	{ cout <<a; }
Val *Expr_num::eval (void) const
	{ return new Val_int (a); };

class Expr_var: public Expr {
	public:
	void print (void) const;
	Val *eval (void) const;
	Val **leval (void) const;
	int islvalue (void) const;
	Expr_var (Xref<Val> *);
	~Expr_var (void);
	private:
	Xref<Val> *const a;
	};
Expr_var::Expr_var (Xref<Val> *u):
	a (u) {
	a->newref ();
	}
void Expr_var::print (void) const
	{ cout <<"v(" <<a <<')'; }
Val *Expr_var::eval (void) const 
	{ return a->get ()->dup (); };
Val **Expr_var::leval (void) const 
	{ return a->adr (); };
int Expr_var::islvalue (void) const
	{ return 1; };
Expr_var::~Expr_var (void) {
	a->delref ();
	}
	
// futasideju scope
struct savevar_t {
	Val **a;
	Val *b;
	savevar_t *c;
	savevar_t (Val **, Val *, savevar_t *);
	};
savevar_t::savevar_t (Val **u, Val *v, savevar_t *w):
	a (u), b(v), c(w) { }
	
stack<savevar_t *> savevar;
void Savevar_init (void)
	{ savevar.push ((savevar_t*)0); }
void Savevar_push (void) 
	{ savevar.push ((savevar_t*)0); }
void Savevar_save (Val **u) {
	savevar.top ()= new savevar_t (u, *u, savevar.top ()); 
	// cout <<"#[DSave]: " <<u <<'\n'; // <---- DEBUG
	}
void Savevar_pop (void) {
	savevar_t *p= savevar.top ();
	savevar.pop ();
	while (p) {
		// cout <<"#[DRstor]: " <<p->a <<'\n'; // <---- DEBUG
		*p->a= p->b;
		p= p->c;
		}
	}

class Expr_newvar: public Expr {
	public:
	void print (void) const;
	Val *eval (void) const;
	Val **leval (void) const;
	int islvalue (void) const;
	Expr_newvar (Xref<Val> *);
	~Expr_newvar (void);
	private:
	Xref<Val> *const a;
	};
Expr_newvar::Expr_newvar (Xref<Val> *u):
	a (u) {
	a->newref ();
	}
void Expr_newvar::print (void) const
	{ cout <<"newvar(" <<a <<')'; }
// Ez egy hibas koncepcio: minden kifet balrol jobbra kell evalni, hogy ez 
// mukodjon.  Elmondhatom szoban, hogy lehetne megjavitani, ha erdekel.
Val *Expr_newvar::eval (void) const {
	return (*leval ())->dup (); 
	};
Val **Expr_newvar::leval (void) const {
	if (*a->adr ())
		Savevar_save (a->adr ());
	*a->adr ()= a->get ()-> dup ();
	return a->adr (); 
	};
int Expr_newvar::islvalue (void) const
	{ return 1; };
Expr_newvar::~Expr_newvar (void) {
	a->delref ();
	}


class Expr_progn: public Expr {
	public:
	void print (void) const;
	Val *eval (void) const;
	~Expr_progn (void);
	Expr_progn (Expr *, Expr ***);
	private:
	Expr *a;
	Expr *b;
	};
void Expr_progn::print (void) const {
	cout <<"progn (";
	a->print ();
	cout <<' ';
	b->print ();
	cout <<")";
	}
Expr_progn::~Expr_progn (void)
	{ delete a; delete b; }
Expr_progn::Expr_progn (Expr *u, Expr ***dp):
	a (u), b (0) { 
	*dp= &b; 
	}
Val *Expr_progn::eval (void) const {
	delete a->eval ();
	return b->eval ();
	}

class Expr_do: public Expr {
	public:
	void print (void) const;
	Val *eval (void) const;
	~Expr_do (void);
	Expr_do (Expr *);
	private:
	Expr *a;
	};
void Expr_do::print (void) const {
	cout <<"do (";
	a->print ();
	cout <<")";
	}
Expr_do::~Expr_do (void)
	{ delete a; }
Expr_do::Expr_do (Expr *u):
	a (u) { }
Val *Expr_do::eval (void) const {
	Val *r= 0;
	r= a->eval ();
	while (r->isa (Val_loop::type)) {
		delete r;
		r= a->eval ();
		}
	return r;
	}

class Expr_lambda: public Expr {
	public:
	void print (void) const;
	Val *eval (void) const;
	Expr_lambda (Expr ***);
	~Expr_lambda (void);
	private:
	Xref<Expr> *const a;
	};
Expr_lambda::Expr_lambda (Expr ***dp):
	a (new Xref<Expr> (dp)) { };
Expr_lambda::~Expr_lambda (void) {
	a->delref ();
	}
Val *Expr_lambda::eval (void) const 
	{ return new Val_lambda (a); }
void Expr_lambda::print (void) const {
	cout <<"lambda (";
	a->get ()->print ();
	cout <<')';
	}

class Expr_and: public Expr {
	public:
	void print (void) const;
	Val *eval (void) const;
	~Expr_and (void);
	Expr_and (Expr *, Expr ***);
	private:
	Expr *a;
	Expr *b;
	};
void Expr_and::print (void) const {
	cout <<"and (";
	a->print ();
	cout <<' ';
	b->print ();
	cout <<")";
	}
Expr_and::~Expr_and (void)
	{ delete a; delete b; }
Expr_and::Expr_and (Expr *u, Expr ***dp):
	a (u), b (0) { 
	*dp= &b; 
	}
Val *Expr_and::eval (void) const {
	Val *t= a->eval ();
	if (!t->isa(Val_int::type)) {
		cerr <<"first argument of operator and is of ";
		fatal (t->isa (Val_void::type) ?
			"void type" : "wrong type");
		}
	if (!((Val_int*)t)->a)
		return t;
	delete t;
	return b->eval ();
	}

class Expr_or: public Expr {
	public:
	void print (void) const;
	Val *eval (void) const;
	~Expr_or (void);
	Expr_or (Expr *, Expr ***);
	private:
	Expr *a;
	Expr *b;
	};
void Expr_or::print (void) const {
	cout <<"or (";
	a->print ();
	cout <<' ';
	b->print ();
	cout <<")";
	}
Expr_or::~Expr_or (void)
	{ delete a; delete b; }
Expr_or::Expr_or (Expr *u, Expr ***dp):
	a (u), b (0) { 
	*dp= &b; 
	}
Val *Expr_or::eval (void) const {
	Val *t= a->eval ();
	if (!t->isa(Val_int::type)) {
		cerr <<"first argument of operator or is of ";
		fatal (t->isa (Val_void::type) ?
			"void type" : "wrong type");
		}
	if (((Val_int*)t)->a)
		return t;
	delete t;
	return b->eval ();
	}

class Expr_and3: public Expr {
	public:
	void print (void) const;
	Val *eval (void) const;
	~Expr_and3 (void);
	Expr_and3 (Expr *, Expr ***, Expr ***);
	private:
	Expr *a;
	Expr *b;
	Expr *c;
	};
void Expr_and3::print (void) const {
	cout <<"and3 (";
	a->print ();
	cout <<' ';
	b->print ();
	cout <<' ';
	c->print ();
	cout <<")";
	}
Expr_and3::~Expr_and3 (void)
	{ delete a; delete b; delete c; }
Expr_and3::Expr_and3 (Expr *u, Expr ***bp, Expr ***cp):
	a (u), b (0), c(0) { 
	*bp= &b; *cp= &c;
	}
Val *Expr_and3::eval (void) const {
	Val *t= a->eval ();
	if (!t->isa(Val_int::type)) {
		cerr <<"first argument of operator and3 is of ";
		fatal (t->isa (Val_void::type) ?
			"void type" : "wrong type");
		}
	int i= ((Val_int*)t)->a;
	delete t;
	if (i)
		return b->eval ();
	else
		return c->eval ();
	}

class Expr_or3: public Expr {
	public:
	void print (void) const;
	Val *eval (void) const;
	~Expr_or3 (void);
	Expr_or3 (Expr *, Expr ***, Expr ***);
	private:
	Expr *a;
	Expr *b;
	Expr *c;
	};
void Expr_or3::print (void) const {
	cout <<"or3 (";
	a->print ();
	cout <<' ';
	b->print ();
	cout <<' ';
	c->print ();
	cout <<")";
	}
Expr_or3::~Expr_or3 (void)
	{ delete a; delete b; delete c; }
Expr_or3::Expr_or3 (Expr *u, Expr ***bp, Expr ***cp):
	a (u), b (0), c(0) { 
	*bp= &b; *cp= &c;
	}
Val *Expr_or3::eval (void) const {
	Val *t= a->eval ();
	if (!t->isa(Val_int::type)) {
		cerr <<"first argument of operator or3 is of ";
		fatal (t->isa (Val_void::type) ?
			"void type" : "wrong type");
		}
	int i= ((Val_int*)t)->a;
	delete t;
	if (!i)
		return b->eval ();
	else
		return c->eval ();
	}

class Expr_let: public Expr {
	public:
	void print (void) const;
	Val *eval (void) const;
	~Expr_let (void);
	Expr_let (Expr *, Expr ***);
	private:
	Expr *a;
	Expr *b;
	};
void Expr_let::print (void) const {
	cout <<"let (";
	a->print ();
	cout <<' ';
	b->print ();
	cout <<")";
	}
Expr_let::~Expr_let (void)
	{ delete a; delete b; }
Expr_let::Expr_let (Expr *u, Expr ***dp):
	a (u), b (0) { 
	*dp= &b; 
	}
Val *Expr_let::eval (void) const {
	Val **t= a->leval ();
	Val *r= b->eval ();
	delete *t;
	*t= r->dup ();
	return r;
	}

class Expr_call: public Expr {
	public:
	Expr_call (Expr *);
	void print (void) const;
	Val *eval (void) const;
	~Expr_call (void);
	private:
	Expr *a;
	};
Expr_call::Expr_call (Expr *u):
	a (u) { }
void Expr_call::print (void) const {
	cout <<"call (";
	a->print ();
	cout <<')';
	}
Expr_call::~Expr_call (void)
	{ delete a; }
Val *Expr_call::eval (void) const {
	Val *r= a->eval ();
	if (!r-> isa(Val_lambda::type))
		fatal ("error: calling non-function");
	Xref<Expr> *t= ((Val_lambda*)r)->get ();
	t-> newref ();
	delete r;
	Savevar_push ();
	r= t->get ()->eval ();
	Savevar_pop ();
	t-> delref ();
	return r; 
	}

template <int n> struct Bfunc {
	const char *const nm;
	Val *(*const op) (Val *[n]);
	const int tm[n?n:1];
	};

template <int n> class Expr_bfunc: public Expr {
	public:
	void print (void) const;
	Val *eval (void) const;
	~Expr_bfunc (void);
	Expr_bfunc (const Bfunc<n> *, Expr *, Expr ***);
	Expr_bfunc (const Bfunc<n> *, Expr ***);
	private:
	const Bfunc<n> *a;
	Expr *b[n?n:1];
	};
template <int n> void Expr_bfunc<n>::print (void) const {
	cout <<a->nm <<" (";
	int i;
	for (i= 0; i<n-1; i++) {
		b[i]->print ();
		cout <<" ";
		}
	if (n>=1)
		b[n-1]->print ();
	cout <<")";
	}
template <int n> Val *Expr_bfunc<n>::eval (void) const {
	int i;
	Val *u[n?n:1];
	for (i= 0; i<n; i++) {
		u[i]= b[i]->eval ();
		if (!u[i]->isa (a->tm[i])) {
			cerr <<"argument nr " <<i+1
				<<" of builtin func " <<a->nm <<" is of ";
			fatal (u[i]->isa (Val_void::type) ?
				"void type" : "wrong type");
			}
		}
	Val *r= a->op (u);
	for (i= 0; i<n; i++) {
		delete u[i];
		}
	return r;
	}
template <int n> Expr_bfunc<n>::~Expr_bfunc (void) {
	int i;
	for (i= 0; i<n; i++)
		delete b[i];
	}
template <int n> Expr_bfunc<n>::Expr_bfunc (
	const Bfunc<n> *f,
	Expr *u,
	Expr ***dp
	):
	a (f)
	{ b[0]= u; if (dp) *dp= 1+b; }
template <int n> Expr_bfunc<n>::Expr_bfunc (const Bfunc<n> *f, Expr ***dp):
	a (f) 
	{ if (dp) *dp= b; }

class Expr_flet: public Expr {
	public:
	void print (void) const;
	Val *eval (void) const;
	~Expr_flet (void);
	Expr_flet (Bfunc<2>*, Expr *, Expr ***);
	private:
	Expr *a;
	Expr *b;
	Bfunc<2> *c;
	};
void Expr_flet::print (void) const {
	cout <<c->nm <<"-let (";
	a->print ();
	cout <<' ';
	b->print ();
	cout <<")";
	}
Expr_flet::~Expr_flet (void)
	{ delete a; delete b; }
Expr_flet::Expr_flet (Bfunc<2> *z, Expr *u, Expr ***dp):
	a (u), b (0), c (z) { 
	*dp= &b; 
	}
Val *Expr_flet::eval (void) const {
	Val *u[2];
	u[0]= a->eval ();
	if (!u[0]->isa (c->tm[0])) {
		cerr <<"first argument of " <<c->nm <<"-let op is of ";
		fatal (u[0]->isa (Val_void::type) ?
			"void type" : "wrong type");
		}
	u[1]= b->eval ();
	if (!u[1]->isa (c->tm[1])) {
		cerr <<"second argument of " <<c->nm <<"-let op is of ";
		fatal (u[1]->isa (Val_void::type) ?
			"void type" : "wrong type");
		}
	Val *r= c->op (u);
	delete u[0]; delete u[1];
	Val **t= a->leval ();
	delete *t;
	*t= r->dup ();
	return r;
	}


Val *Bfunc_add_op (Val *u[2])
	{ return new Val_int (((Val_int*)u[0])->a+((Val_int*)u[1])->a); }
Bfunc<2> Bfunc_add= 
	{"add", &Bfunc_add_op, {Val_int::type, Val_int::type}};
Val *Bfunc_sub_op (Val *u[2])
	{ return new Val_int (((Val_int*)u[0])->a-((Val_int*)u[1])->a); }
Bfunc<2> Bfunc_sub= 
	{"sub", &Bfunc_sub_op, {Val_int::type, Val_int::type}};
Val *Bfunc_mul_op (Val *u[2])
	{ return new Val_int (((Val_int*)u[0])->a*((Val_int*)u[1])->a); }
Bfunc<2> Bfunc_mul= 
	{"mul", &Bfunc_mul_op, {Val_int::type, Val_int::type}};
Val *Bfunc_div_op (Val *u[2]) {
	long i1= ((Val_int*)u[0])->a;
	long i2= ((Val_int*)u[1])->a;
	if (i2==0 || i1==LONG_MIN && i2==-1)
		fatal ("division by zero error");
	return new Val_int (i1/i2);
	}
Bfunc<2> Bfunc_div= 
	{"div", &Bfunc_div_op, {Val_int::type, Val_int::type}};
Val *Bfunc_mod_op (Val *u[2]) {
	long i1= ((Val_int*)u[0])->a;
	long i2= ((Val_int*)u[1])->a;
	if (i2==0 || i1==LONG_MIN && i2==-1)
		fatal ("modulo by zero error");
	return new Val_int (i1%i2);
	}
Bfunc<2> Bfunc_mod= 
	{"mod", &Bfunc_mod_op, {Val_int::type, Val_int::type}};

Val *Bfunc_min_op (Val *u[2]) {
	return new Val_int (
		min (((Val_int*)u[0])->a, ((Val_int*)u[1])->a)
		); 
	}
Bfunc<2> Bfunc_min= 
	{"min", &Bfunc_min_op, {Val_int::type, Val_int::type}};
Val *Bfunc_max_op (Val *u[2]) {
	return new Val_int (
		max (((Val_int*)u[0])->a, ((Val_int*)u[1])->a)
		);
	}
Bfunc<2> Bfunc_max= 
	{"max", &Bfunc_max_op, {Val_int::type, Val_int::type}};

Val *Bfunc_eq_op (Val *u[2])
	{ return new Val_int (((Val_int*)u[0])->a==((Val_int*)u[1])->a); }
Bfunc<2> Bfunc_eq= 
	{"eq", &Bfunc_eq_op, {Val_int::type, Val_int::type}};
Val *Bfunc_ne_op (Val *u[2])
	{ return new Val_int (((Val_int*)u[0])->a!=((Val_int*)u[1])->a); }
Bfunc<2> Bfunc_ne= 
	{"ne", &Bfunc_ne_op, {Val_int::type, Val_int::type}};
Val *Bfunc_lt_op (Val *u[2])
	{ return new Val_int (((Val_int*)u[0])->a<((Val_int*)u[1])->a); }
Bfunc<2> Bfunc_lt= 
	{"lt", &Bfunc_lt_op, {Val_int::type, Val_int::type}};
Val *Bfunc_gt_op (Val *u[2])
	{ return new Val_int (((Val_int*)u[0])->a>((Val_int*)u[1])->a); }
Bfunc<2> Bfunc_gt= 
	{"gt", &Bfunc_gt_op, {Val_int::type, Val_int::type}};
Val *Bfunc_le_op (Val *u[2])
	{ return new Val_int (((Val_int*)u[0])->a<=((Val_int*)u[1])->a); }
Bfunc<2> Bfunc_le= 
	{"le", &Bfunc_le_op, {Val_int::type, Val_int::type}};
Val *Bfunc_ge_op (Val *u[2])
	{ return new Val_int (((Val_int*)u[0])->a>=((Val_int*)u[1])->a); }
Bfunc<2> Bfunc_ge= 
	{"ge", &Bfunc_ge_op, {Val_int::type, Val_int::type}};

Val *Bfunc_neg_op (Val *u[1])
	{ return new Val_int (-((Val_int*)u[0])->a); }
Bfunc<1> Bfunc_neg=
	{"neg", &Bfunc_neg_op, {Val_int::type}};
Val *Bfunc_not_op (Val *u[1])
	{ return new Val_int (!((Val_int*)u[0])->a); }
Bfunc<1> Bfunc_not=
	{"not", &Bfunc_not_op, {Val_int::type}};

template <int m> Val *Bfunc_isa_op (Val *u[1])
	{ return new Val_int (u[0]->isa (m)!=0); }
Bfunc<1> Bfunc_voidp=
	{"voidp", &Bfunc_isa_op<Val_void::type>, {~0}};
Bfunc<1> Bfunc_intp=
	{"intp", &Bfunc_isa_op<Val_int::type>, {~0}};
Bfunc<1> Bfunc_lambdap=
	{"lambdap", &Bfunc_isa_op<Val_lambda::type>, {~0}};

Val *Bfunc_abs_op (Val *u[1]) 
	{ return new Val_int (abs(((Val_int*)u[0])->a)); }
Bfunc<1> Bfunc_abs=
	{"abs", &Bfunc_abs_op, {Val_int::type}};
Val *Bfunc_rand_op (Val *u[1]) {
	return new Val_int (static_cast<int>(
		(double)((Val_int*)u[0])->a*rand ()/((double)RAND_MAX+1)
		)); 
	}
Bfunc<1> Bfunc_rand=
	{"rand", &Bfunc_rand_op, {Val_int::type}};

Val *Bfunc_print_op (Val *u[1]) {
	if (u[0]->isa (Val_int::type)) 
		cout <<((Val_int*)u[0])->a <<'\n';
	return new Val_void;
	}
Bfunc<1> Bfunc_print=
	{"print", &Bfunc_print_op, {~0}};

Val *Bfunc_exit_op (Val *u[0])
	{ exit (0); }
Bfunc<0> Bfunc_exit=
	{"exit", &Bfunc_exit_op, {}};



// Symbol -----------------------------------------------------

class Symbol {
	public:
	virtual ~Symbol (void);
	virtual int isa (int) const= 0;
	protected:
	Symbol (void);
	};

class Symb_reserved: public Symbol {
	public:
	Symb_reserved (int);
	int isa (int m) const { return m&kind; }
	static const int kind= 16;
	int get (void) const;
	private:
	int n;
	};

class Symb_var: public Symbol {
	public:
	Symb_var (void);
	int isa (int m) const { return m&kind; }
	static const int kind= 1;
	Xref<Val> *get (void);
	~Symb_var (void);
	private:
	Xref<Val> *a;
	};

template <int n> class Symb_bfunc: public Symbol {
	public:
	Symb_bfunc (Bfunc<n> *);
	int isa (int m) const { return m&kind; }
	static const int kind= 
		n==1 ? 2 :
		n==0 ? 4 :
		n==2 ? 8 : 0;
	Bfunc<n> const *get (void) const;
	private:
	Bfunc<n> *const a;
	};
	

Symbol::Symbol (void)
	{ }
Symbol::~Symbol (void)
	{ }
int Symbol::isa (int m) const
	{ fatal ("internal error: abstract Symbol"); return 0; }

Symb_reserved::Symb_reserved (int u):
	n (u) { };
int Symb_reserved::get (void) const
	{ return n; }

Symb_var::Symb_var (void):
	a (new Xref<Val> (new Val_void)) { }
/*
Symb_var::Symb_var (void):
	a (new Xref<Val> ((Val*)0)) { }
*/
Xref<Val> *Symb_var::get (void)
	{ return a; }
Symb_var::~Symb_var (void) 
	{ a->delref (); }

template <int n> Symb_bfunc<n>::Symb_bfunc (Bfunc<n> *u):
	a (u) {
	if (!kind) 
		fatal ("internal error: invalid Symb_bfunc"); 
	}
template <int n> Bfunc<n> const *Symb_bfunc<n>::get (void) const
	{ return a; }



// symbt ------------------------------------------------------

enum { // reserved words
	R_VAR= 1
	};


typedef map <string, Symbol *> symbt_t;
typedef symbt_t::iterator symbt_it;

symbt_t symbt;

struct lscope_listt {
 	string a;
	Symbol *c;
	lscope_listt *d;
	int b;
	lscope_listt (string&, Symbol *w, lscope_listt *);
	};

int lscope;
lscope_listt *lscopel= 0;


Symbol *Symbt_get (string s) {
	symbt_it f= symbt.find (tokenval_s);
	if (f==symbt.end ())
		return 0;
	else
		return f->second;
	}
lscope_listt::lscope_listt (string& u, Symbol *w, lscope_listt *v):
	a (u), c(w), d (v), b (lscope) { }
void Symbt_add (string s, Symbol *v) {
	symbt_it f= symbt.find (tokenval_s);
	Symbol *o;
	if (f==symbt.end ())
		o= 0;
	else {
		if (!f->second->isa(Symb_var::kind))
			fatal ("error: trying to redefine builtin symbol");
		o= f->second;
		}
	symbt[s]= v;	
	lscope_listt *u= lscopel;
	lscopel= new lscope_listt (s, o, u);
	}
void Lscope_push (void) {
	lscope++;
	if (lscope<0)
		fatal ("internal error panic: lscope number overflow, this is impossible");
	}
void Lscope_pop (void) {
	symbt_it f;
	while (lscopel->b>=lscope) {
		f= symbt.find (lscopel->a);
		delete f->second;
		if (lscopel->c)
			f->second= lscopel->c;
		else
			symbt.erase (f);
		lscopel= lscopel->d;
		}
	lscope--;
	}


void symbt_init (void) {
	Symbt_add ("var", new Symb_reserved (R_VAR));
	Symbt_add ("print", new Symb_bfunc<1> (&Bfunc_print));
	Symbt_add ("abs", new Symb_bfunc<1> (&Bfunc_abs));
	Symbt_add ("rand", new Symb_bfunc<1> (&Bfunc_rand));
	Symbt_add ("exit", new Symb_bfunc<0> (&Bfunc_exit));
	Symbt_add ("min", new Symb_bfunc<2> (&Bfunc_min));
	Symbt_add ("max", new Symb_bfunc<2> (&Bfunc_max));
	Symbt_add ("voidp", new Symb_bfunc<1> (&Bfunc_voidp));
	Symbt_add ("intp", new Symb_bfunc<1> (&Bfunc_intp));
	Symbt_add ("lambdap", new Symb_bfunc<1> (&Bfunc_lambdap));
	char c;
	if (interactive&INTERA_TTY)
		for (c= 'a'; c<='z'; c++) {
			// EBCDIC charsetben sok nem-betu is bekerul a
			// symbt-be, de ezeket sosem fogjuk elerni, mert
			// a flexer nem ad ilyen azonositot.
			Symbt_add (string (1, c), new Symb_var);
			}
	};


// Parse ------------------------------------------------------ 

int interactive= 0;

int tokbufu= 0;
int tokbuf= -1;
int gettok (void) {
	if (tokbufu) {
		tokbufu= 0;
		return tokbuf;
		}
	else
		return tokbuf= yylex ();
	};
void ungettok (void) {
	tokbufu= 1;
	};
void ungettok (int t) {
	if (tokbufu)
		fatal ("internal error: cannot unget more then one token");
	tokbufu= 1;
	tokbuf= t;
	};


// Ez egy nagy es ormotlan fuggveny.
void Parse_expr (Expr **r, int h) {
	Expr **p;
	Expr **q;
	int lowpr= 0;
	int t= gettok ();
	if (t==T_INT)
		*r= new Expr_num (tokenval_n);
	else if (t==T_ID) {
		Symbol *s= Symbt_get (tokenval_s);
		if (!s) {
			cerr <<"error: symbol " <<tokenval_s;
			fatal (" does not exist");
			}
		else if (s->isa (Symb_bfunc<1>::kind)) {
			*r= new Expr_bfunc<1> (
				((Symb_bfunc<1>*)s)->get (), &p
				);
			if (gettok ()!=T_OPEN)
				fatal ("parse error: '(' expected");
			Parse_expr (p, 240);
			if (gettok ()!=T_CLOSE)
				fatal ("parse error: ')' expected");
			}
		else if (s->isa (Symb_bfunc<2>::kind)) {
			*r= new Expr_bfunc<2> (
				((Symb_bfunc<2>*)s)->get (), &p
				);
			if (gettok ()!=T_OPEN)
				fatal ("parse error: '(' expected");
			Parse_expr (p, 240);
			if (gettok ()!=T_COMMA)
				fatal ("parse error: ',' expected");
			Parse_expr (p+1, 240);
			if (gettok ()!=T_CLOSE)
				fatal ("parse error: ')' expected");
			}
		else if (s->isa (Symb_bfunc<0>::kind)) {
			*r= new Expr_bfunc<0> (
				((Symb_bfunc<0>*)s)->get (), 0
				);
			if (gettok ()==T_OPEN) {
				if (gettok ()!=T_CLOSE)
					fatal ("parse error: ')' expected");
				} // tanulsag: ide KELL a kapcsos
			else
				ungettok ();
			}
		else if (s->isa (Symb_var::kind)) {
			*r= new Expr_var (((Symb_var*)s)->get ());
			}
		else if (s->isa (Symb_reserved::kind)) {
		int n= ((Symb_reserved*)s)->get ();
		if (n==R_VAR) {
			if (gettok ()!=T_ID)
				fatal ("parse error: id expected after 'var'");
			Symb_var *v= new Symb_var;
			Symbt_add (tokenval_s, v);
			*r= new Expr_newvar (v->get ());
			}
		else
			fatal ("internal error: unknown Symb_reserved");
		} // Symb_reserved
		else
			fatal ("internal error: parsing unknown kind of Symbol");			
		} // T_ID
	else if (t==T_MINUS) {
		*r= new Expr_bfunc<1> (&Bfunc_neg, &p);
		Parse_expr (p, min (h, 49));
		}
	else if (t==T_HAT) {
		*r= new Expr_bfunc<1> (&Bfunc_not, &p);
		Parse_expr (p, min (h, 140));
		}
	else if (t==T_PLUS)
		Parse_expr (r, min (h, 49));
	else if (t==T_OPEN) {
		Lscope_push ();
		Parse_expr (r, 300);
		t= gettok ();
		if (t==T__LOOP) {
			*r= new Expr_do (*r);
			t= gettok ();
			}
		if (t!=T_CLOSE)
			fatal (/*(*/"parse error: ')' expected");
		Lscope_pop ();
		}
	else if (t==T_BOPEN) {
		Lscope_push ();
		*r= new Expr_lambda (&p); 
		Parse_expr (p, 300);
		t= gettok ();
		if (t==T__LOOP) {
			*p= new Expr_do (*p);
			t= gettok ();
			}
		if (t!=T_BCLOSE)
			fatal (/*(*/"parse error: '}' expected");
		Lscope_pop ();
		}
	else if (t==T_STAR && h>=270) {
		*r= new Expr_loop;
		ungettok (T__LOOP);
		return;
		}
	else if (h>=250) {
		*r= new Expr_void;
		ungettok ();
		lowpr= 1;
		}
	else
		fatal ("parse error: expr expected");
	do {
	t= gettok ();
	if (t==T_SC && h>=270) {
		*r= new Expr_progn (*r, &p);
		if (h==270)
			return Parse_expr (p, 270);
		Parse_expr (p, 270);
		}
	else if (t==T_QUES && h>=270) {
		*r= new Expr_and3 (*r, &p, &q);
		Parse_expr (p, 260);
		if (gettok ()!=T_SC)
			fatal ("parse error: ';' expected to close '?' expr");
		if (h==270)
			return Parse_expr (q, 270);
		Parse_expr (q, 270);
		}
	else if (t==T_BANG && h>=270) {
		*r= new Expr_or3 (*r, &p, &q);
		Parse_expr (p, 260);
		if (gettok ()!=T_SC)
			fatal ("parse error: ';' expected to close '!' expr");
		if (h==270)
			return Parse_expr (q, 270);
		Parse_expr (q, 270);
		}
	else if (t==T_COLON && h>=260) {
		*r= new Expr_progn (*r, &p);
		if (h==260)
			return Parse_expr (p, 260);
		Parse_expr (p, 260);
		} 
	else if (lowpr) {
		ungettok ();
		return;
		}
	else if (t==T_BANGBANG && h>=190) {
		*r= new Expr_or (*r, &p);
		Parse_expr (p, 190);
		} 
	else if (t==T_QUESQUES && h>=190) {
		*r= new Expr_and (*r, &p);
		Parse_expr (p, 190);
		} 
	else if (t==T_COLONCOLON && h>=180) {
		*r= new Expr_progn (*r, &p);
		Parse_expr (p, 180);
		} 
	else if (t==T_LET && h>=170) {
		if (!(*r)->islvalue ())
			fatal ("parse error: assignment to non-lvalue");
		*r= new Expr_let (*r, &p);
		Parse_expr (p, 170);
		}
	else if (t==T_PLUSLET && h>=170) {
		if (!(*r)->islvalue ())
			fatal ("parse error: add-assignment to non-lvalue");
		*r= new Expr_flet (&Bfunc_add, *r, &p);
		Parse_expr (p, 170);
		}
	else if (t==T_MINUSLET && h>=170) {
		if (!(*r)->islvalue ())
			fatal ("parse error: sub-assignment to non-lvalue");
		*r= new Expr_flet (&Bfunc_sub, *r, &p);
		Parse_expr (p, 170);
		}
	else if (t==T_STARLET && h>=170) {
		if (!(*r)->islvalue ())
			fatal ("parse error: mul-assignment to non-lvalue");
		*r= new Expr_flet (&Bfunc_mul, *r, &p);
		Parse_expr (p, 170);
		}
	else if (t==T_SLASHLET && h>=170) { // a '/' hivatalos neve 'solidus'
		if (!(*r)->islvalue ())
			fatal ("parse error: div-assignment to non-lvalue");
		*r= new Expr_flet (&Bfunc_div, *r, &p);
		Parse_expr (p, 170);
		}
	else if (t==T_PERCENTLET && h>=170) {
		if (!(*r)->islvalue ())
			fatal ("parse error: mod-assignment to non-lvalue");
		*r= new Expr_flet (&Bfunc_mod, *r, &p);
		Parse_expr (p, 170);
		}
	else if (t==T_BARBAR && h>=160) {
		*r= new Expr_or (*r, &p);
		Parse_expr (p, 160);
		} 
	else if (t==T_AMPAMP && h>=150) {
		*r= new Expr_and (*r, &p);
		Parse_expr (p, 150);
		} 
	else if (t==T_EQ && h>=130) {
		*r= new Expr_bfunc<2> (&Bfunc_eq, *r, &p);
		Parse_expr (p, 130);
		} 
	else if (t==T_NE && h>=130) {
		*r= new Expr_bfunc<2> (&Bfunc_ne, *r, &p);
		Parse_expr (p, 130);
		} 
	else if (t==T_LT && h>=130) {
		*r= new Expr_bfunc<2> (&Bfunc_lt, *r, &p);
		Parse_expr (p, 130);
		} 
	else if (t==T_GT && h>=130) {
		*r= new Expr_bfunc<2> (&Bfunc_gt, *r, &p);
		Parse_expr (p, 130);
		} 
	else if (t==T_LE && h>=130) {
		*r= new Expr_bfunc<2> (&Bfunc_le, *r, &p);
		Parse_expr (p, 130);
		} 
	else if (t==T_GE && h>=130) {
		*r= new Expr_bfunc<2> (&Bfunc_ge, *r, &p);
		Parse_expr (p, 130);
		} 
	else if (t==T_PLUS && h>=50) {
		*r= new Expr_bfunc<2> (&Bfunc_add, *r, &p);
		Parse_expr (p, 49);
		}
	else if (t==T_MINUS && h>=50) {
		*r= new Expr_bfunc<2> (&Bfunc_sub, *r, &p);
		Parse_expr (p, 49);
		}
	else if (t==T_STAR && h>=40) {
		*r= new Expr_bfunc<2> (&Bfunc_mul, *r, &p);
		Parse_expr (p, 39);
		}
	else if (t==T_SLASH && h>=40) {
		*r= new Expr_bfunc<2> (&Bfunc_div, *r, &p);
		Parse_expr (p, 39);
		}
	else if (t==T_PERCENT && h>=40) {
		*r= new Expr_bfunc<2> (&Bfunc_mod, *r, &p);
		Parse_expr (p, 39);
		}
	else if (t==T_OPEN && h>=30) {
		*r= new Expr_call (*r);
		if (gettok ()!=T_CLOSE)
			fatal ("parse error: ')' expected to close function call");
		}
	else {
		ungettok ();
		return;
		}
	}
	while (1);
	}


// Igazad volt, ezt igy tenyleg nem tudom megcsinalni szepre.
void Parse_cmd1 (Expr **p) {
	ungettok ();
	Parse_expr (p, 250);
	*p= new Expr_bfunc<1> (&Bfunc_print, *p, 0);
	if (gettok ()!=T_SC)
		fatal ("parse error: ';' expected");
	interactive|= INTERA_NL;
	if (gettok ()!=T_NL) {
		interactive &= ~INTERA_NL;
		Expr **dp;
		*p= new Expr_progn (*p, &dp);
		Parse_cmd1 (dp);
		}
	};

Expr *Parse_cmd (void) {
	interactive|= INTERA_NL;
	int t;
	do
		cout <<"s> ";
	while (T_NL==(t= gettok ()));
	if (t==T_EOF)
		return 0;
	interactive&= ~INTERA_NL;
	Expr *p;
	Parse_cmd1 (&p);
	return p;
	};
	
Expr *Parse_prog (void) {
	Expr *p;
	Parse_expr (&p, 300);
	return p;
	};


// main ------------------------------------------------------------

int main (void) {
	Expr *t;
	if (isatty (0))
		interactive= INTERA_TTY;
	srand (time (0)); // valojaban jobb a /dev/urandom-bol szerezni
	// random-seedet, mert igy egy masdpercen belul ketszer elinditva
	// a progit, ugyanaz lesz a random seed.
	symbt_init ();
	Savevar_init ();
	if (interactive&INTERA_TTY) {
 		while (0!=(t= Parse_cmd ())) {
			cout <<"# "; // <---- DEBUG
			t->print ();
			cout <<";\n";
			delete t->eval ();
			delete t;
			}
		cout <<'\n';
		}
	else {
		t= Parse_prog ();
		cout <<"# "; // <---- DEBUG
		t->print ();
		cout <<";\n";
		delete t->eval ();
		delete t;
		}
	return 0;
	}

// second pass --------------------------------------------------

Val_lambda::~Val_lambda (void)
	{ a->delref (); }



//__END__
